/* Substring test for UTF-8/UTF-16/UTF-32 strings.
   Copyright (C) 2002, 2006, 2009-2017 Free Software Foundation, Inc.
   Written by Bruno Haible <bruno@clisp.org>, 2002.

   This program is free software: you can redistribute it and/or
   modify it under the terms of either:

     * the GNU Lesser General Public License as published by the Free
       Software Foundation; either version 3 of the License, or (at your
       option) any later version.

   or

     * the GNU General Public License as published by the Free
       Software Foundation; either version 2 of the License, or (at your
       option) any later version.

   or both in parallel, as here.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

bool
FUNC (const UNIT *str, const UNIT *suffix)
{
  size_t len = U_STRLEN (str);
  size_t suffixlen = U_STRLEN (suffix);

  if (len >= suffixlen)
    return (U_CMP (str + (len - suffixlen), suffix, suffixlen) == 0);
  else
    return false;
}
