/* DO NOT EDIT! GENERATED AUTOMATICALLY! */
/* Properties of Unicode characters.  */
/* Generated automatically by gen-uni-tables.c for Unicode 9.0.0.  */
#define header_0 16
#define header_2 9
#define header_3 127
#define header_4 15
static const
struct
  {
    int header[1];
    int level1[3];
    short level2[3 << 7];
    /*unsigned*/ int level3[13 << 4];
  }
u_property_ideographic =
{
  { 3 },
  {
        4 * sizeof (int) / sizeof (short) +     0,
        4 * sizeof (int) / sizeof (short) +   128,
        4 * sizeof (int) / sizeof (short) +   256
  },
  {
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +     0,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    32,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    48,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +    64,
        4 +   384 * sizeof (short) / sizeof (int) +    80,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    96,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +   112,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +   128,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +   144,
        4 +   384 * sizeof (short) / sizeof (int) +   160,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +   176,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
       -1,
        4 +   384 * sizeof (short) / sizeof (int) +    16,
        4 +   384 * sizeof (short) / sizeof (int) +   192,
       -1,
       -1
  },
  {
    0x000000C0, 0x070003FE, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x003FFFFF, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x003FFFFF, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF3FFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x03FFFFFF, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00001FFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x0007FFFF,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0x007FFFFF, 0x00000000,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x001FFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0x3FFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
    0xFFFFFFFF, 0x00000003, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x3FFFFFFF, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
  }
};
